/**
 * Copyright © 2014-2019 Darwino, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.darwino.domino.napi.debug;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Analyzes the memory allocation from a console file
 * 
 * @author priand
 */
public class MemoryAllocAnalyzer {
	
	// https://www.regexplanet.com/advanced/java/index.html

	// [0234:0014-17AC] OSLocalAllocNZ: Allocated 15 (3 + 12) at 0x144701F8
	// [0234:0014-17AC] OSLocalAlloc: Allocated 15 (3 + 12) at 0x144701F8
	// \[(.*)\].*OSLocalAlloc.*at (0x[0-9A-Fa-f]{1,8})
	private static final Pattern regexAlloc = Pattern.compile("\\[(.*)\\].*OSLocalAlloc.*at (0x[0-9A-Fa-f]{1,8})");
	
	// [0234:0014-17AC] OSLocalFree: Freeing 0x144701F8 of size 3
	// \[(.*)\].*OSLocalFree.* (0x[0-9A-Fa-f]{1,8})
	private static final Pattern regexFree = Pattern.compile("\\[(.*)\\].*OSLocalFree.* (0x[0-9A-Fa-f]{1,8})");
	
	// [1F7C:0014-1BC8]   61819F7F [load addr 0x19180000]   j9vm28! VMprJavaSendNative +0x65f, jnisend.asm line 580
	// \[(.*)\].*\[load addr.*
	private static final Pattern regexLoadAddr = Pattern.compile("\\[(.*)\\].*\\[load addr.*");
	
	private class ThreadData {
		String pointer;
		Map<String,StringBuilder> pointers = new HashMap<String, StringBuilder>();
	}
	
	private File file;
	private File output;
	private Map<String,ThreadData> alloc = new HashMap<String,ThreadData>();
	
	public MemoryAllocAnalyzer(File file, File output) {
		this.file = file;
		this.output = output;
	}
	
	public void run() throws Exception {
		PrintStream ps = null;
		if(output!=null) {
			ps = new PrintStream(new FileOutputStream(output));
		}
		try {
			processFile();
			dumpLeaks(ps);
		} finally {
			if(ps!=null) {
				ps.close();
			}
		}
	}
	
	public void processFile() throws Exception {
		BufferedReader r = new BufferedReader(new FileReader(file));
		try {
			do {
				String line = r.readLine();
				if(line==null) {
					break;
				}
				if(line.contains("[load addr")) {
					// 5F99AB0E [load addr 0x168b0000]   nnotes! OSLocalAllocNZ@8 +0x19e
					Matcher m = regexLoadAddr.matcher(line);
					if(m.find()) {
						String thread = m.group(1);
						ThreadData data = alloc.get(thread);
						// Exclude the pbackend classes!
						if(data!=null && data.pointer!=null) { // Should be non null!
							if(line.contains("nlsxbe!")) {
								data.pointers.remove(data.pointer);
							} else {
								StringBuilder b = data.pointers.get(data.pointer);
								if(b!=null) {
									b.append(line);
									b.append('\n');
									continue;
								}
							}
						}
					}
				}
				if(line.contains("OSLocalAlloc")) {
					Matcher m = regexAlloc.matcher(line);
					if(m.find()) {
						String thread = m.group(1);
						String pointer = m.group(2);
						//println("OSLocalAlloc Thread: ["+thread+"]="+pointer);
						if(!alloc.containsKey(thread)) {
							alloc.put(thread, new ThreadData());
						}
						ThreadData data = alloc.get(thread);
						data.pointer = pointer;
						StringBuilder b = new StringBuilder();
						b.append(line);
						b.append('\n');
						data.pointers.put(pointer,b);
					}
				}
				if(line.contains("OSLocalFree")) {
					Matcher m = regexFree.matcher(line);
					if(m.find()) {
						String thread = m.group(1);
						String pointer = m.group(2);
						ThreadData data = alloc.get(thread);
						if(data!=null) {
							//println("OSLocalFree Thread: ["+thread+"]="+pointer);
							data.pointers.remove(pointer);
							data.pointer = null;
						}
					}
				}
			} while(true);
		} finally {
			r.close();
		}
	}
	
	public void dumpLeaks(PrintStream ps) throws Exception {
		println(ps,"--------------------------------");

		int leaks = 0;
		for(Map.Entry<String, ThreadData> e: alloc.entrySet()) {
			leaks += e.getValue().pointers.size();
		}
		println(ps,"Memory Leaks: #"+leaks);
		
		int i=0;
		for(Map.Entry<String, ThreadData> e: alloc.entrySet()) {
			for(Map.Entry<String, StringBuilder> e2: e.getValue().pointers.entrySet()) {
				println(ps,"----- ["+(i++)+"]="+e2.getKey()+" ("+e.getKey()+")");
				println(ps,e2.getValue().toString());
			}
		}
	}
	
	public void println(PrintStream ps, String msg) {
		if(ps!=null) {
			ps.println(msg);
		}
		System.out.println(msg);
	}
	

	public static void main(String[] args) {
		try {
			String path = "c:\\temp\\console.log";
			if(args.length>=1) {
				path = args[0];
			}

			String alloc = "c:\\temp\\alloc.txt";
			if(args.length>=2) {
				alloc = args[1];
			}
			
			MemoryAllocAnalyzer a = new MemoryAllocAnalyzer(new File(path), new File(alloc));
			a.run();
		} catch(Throwable t) {
			t.printStackTrace();
		}
	}

}
