/**
 * Copyright © 2014-2019 Darwino, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.darwino.domino.napi.struct.cd;

import java.util.Collection;
import java.util.Set;

import com.darwino.domino.napi.DominoAPI;
import com.darwino.domino.napi.c.C;
import com.darwino.domino.napi.enums.ActionFlag;
import com.darwino.domino.napi.enums.ActionType;
import com.darwino.domino.napi.enums.DominoEnumUtil;
import com.darwino.domino.napi.enums.ODSType;
import com.darwino.domino.napi.struct.SMM;
import com.darwino.domino.napi.wrap.NSFFormula;
import com.ibm.commons.util.StringUtil;

/**
 * (editods.h)
 * 
 * @author Jesse Gallagher
 * @since 0.8.0
 *
 */
public class CDACTION extends BaseCDStruct<LSIG> {
	
	static {
		int[] sizes = new int[8];
		initNative(sizes);
		sizeOf = sizes[0];
		_Header = sizes[1];
		_Type = sizes[2];
		_IconIndex = sizes[3];
		_Flags = sizes[4];
		_TitleLen = sizes[5];
		_FormulaLen = sizes[6];
		_ShareId = sizes[7];
	}
	private static final native void initNative(int[] sizes);

	
	// Statically initialized fields
	public static final int sizeOf;
	public static final int _Header;
	public static final int _Type;
	public static final int _IconIndex;
	public static final int _Flags;
	public static final int _TitleLen;
	public static final int _FormulaLen;
	public static final int _ShareId;
	

	// Structure-related fields stored internally due to canonical format
	private String title;
	private byte[] actionData;
	private byte[] hideWhenFormula;
	
	public CDACTION() {
		super(SMM.malloc(sizeOf),true);
	}
	public CDACTION(long data) {
		super(data, false);
	}
	public CDACTION(long data, boolean owned) {
		super(data, owned);
	}
	
	/* (non-Javadoc)
	 * @see com.darwino.domino.napi.struct.ODSStruct#getODSType()
	 */
	@Override
	public ODSType getODSType() {
		return ODSType.CDACTION;
	}
	
	/* (non-Javadoc)
	 * @see com.darwino.domino.napi.struct.cd.CDStruct#loadVariableData(com.darwino.domino.napi.DominoAPI, long)
	 */
	@Override
	protected void _readODSVariableData(DominoAPI api, long dataPtr) {
		long ptr = dataPtr;
		
		// Title
		int titleLen = getTitleLen() & 0xFFFFFFFF;
		this.title = C.getLMBCSString(ptr, 0, titleLen);
		ptr = C.ptrAdd(ptr, titleLen + (1 & titleLen));
		
		// Action data
		int formulaLen = getFormulaLen() & 0xFFFFFFFF;
		int actionLen = getHeader().getLength() - api.ODSLength(this.getODSType().getValue()) - titleLen - (1 & titleLen) - formulaLen - (1 & formulaLen);
		this.actionData = new byte[actionLen];
		C.readByteArray(this.actionData, 0, ptr, 0, actionLen);
		ptr = C.ptrAdd(ptr, actionLen);
		
		// Hide-when formula
		this.hideWhenFormula = new byte[formulaLen];
		C.readByteArray(hideWhenFormula, 0, ptr, 0, formulaLen);
	}
	
	/* ******************************************************************************
	 * Struct field getters/setters
	 ********************************************************************************/

	@Override
	public LSIG getHeader() {
		_checkRefValidity();
		return new LSIG(C.ptrAdd(data, _Header));
	}
	@Override
	public void setHeader(LSIG signature) {
		_checkRefValidity();
		C.memcpy(data, _Header, signature.getDataPtr(), 0, LSIG.sizeOf);
	}
	
	public short getTypeRaw() {
		return _getWORD(_Type);
	}
	public void setTypeRaw(short type) {
		_setWORD(_Type, type);
	}
	
	public short getIconIndex() {
		return _getWORD(_IconIndex);
	}
	public void setIconIndex(short iconIndex) {
		_setWORD(_IconIndex, iconIndex);
	}
	
	public int getFlagsRaw() {
		return _getDWORD(_Flags);
	}
	
	public void setFlagsRaw(int flags) {
		_setDWORD(_Flags, flags);
	}
	
	public short getTitleLen() {
		return _getWORD(_TitleLen);
	}
	public void setTitleLen(short titleLen) {
		_setWORD(_TitleLen, titleLen);
	}
	
	public short getFormulaLen() {
		return _getWORD(_FormulaLen);
	}
	public void setFormulaLen(short formulaLen) {
		_setWORD(_FormulaLen, formulaLen);
	}
	
	public int getShareId() {
		return _getDWORD(_ShareId);
	}
	public void setShareId(int shareId) {
		_setDWORD(_ShareId, shareId);
	}
	
	// *******************************************************************************
	// * Encapsulated getters/setters
	// *******************************************************************************

	public Set<ActionFlag> getFlags() {
		return DominoEnumUtil.valuesOf(ActionFlag.class, getFlagsRaw());
	}
	
	public void setFlags(Collection<ActionFlag> flags) {
		setFlagsRaw(DominoEnumUtil.toBitField(ActionFlag.class, flags));
	}
	
	public ActionType getType() {
		return DominoEnumUtil.valueOf(ActionType.class, getTypeRaw());
	}
	
	public void setType(ActionType type) {
		setTypeRaw(type.getValue());
	}
	
	public NSFFormula getHideWhenFormula(DominoAPI api) {
		if(this.hideWhenFormula.length == 0) {
			return null;
		} else {
			return new NSFFormula(api, this.hideWhenFormula);
		}
	}
	
	public byte[] getActionData() {
		return this.actionData;
	}
	
	/* ******************************************************************************
	 * Variable data
	 ********************************************************************************/
	
	public String getTitle() {
		return title;
	}
	
	/* ******************************************************************************
	 * Misc.
	 ********************************************************************************/
	
	/* (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString() {
    	if(isRefValid()) {
	    	return StringUtil.format("[{0}: Header={1}, Text={2}]", //$NON-NLS-1$
	    			getClass().getSimpleName(),
	    			getHeader(),
	    			getTitle()
	    	);
    	} else {
    		return super.toString();
    	}
    }

}
