/**
 * Copyright © 2014-2019 Darwino, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.darwino.domino.napi.struct.cd;

import java.util.Collection;
import java.util.Set;

import com.darwino.domino.napi.DominoAPI;
import com.darwino.domino.napi.c.C;
import com.darwino.domino.napi.enums.BARREC;
import com.darwino.domino.napi.enums.DominoEnumUtil;
import com.darwino.domino.napi.enums.ODSType;
import com.darwino.domino.napi.struct.SMM;
import com.ibm.commons.util.StringUtil;

/**
 * (editods.h)
 * 
 * @author Jesse Gallagher
 * @since 0.8.0
 *
 */
public class CDBAR extends BaseCDStruct<WSIG> {
	
	static {
		int[] sizes = new int[4];
		initNative(sizes);
		sizeOf = sizes[0];
		_Header = sizes[1];
		_Flags = sizes[2];
		_FontID = sizes[3];
	}
	private static final native void initNative(int[] sizes);

	
	// Statically initialized fields
	public static final int sizeOf;
	public static final int _Header;
	public static final int _Flags;
	public static final int _FontID;
	

	// Structure-related fields stored internally due to canonical format
	private String text;
	private short color;
	
	public CDBAR() {
		super(SMM.malloc(sizeOf),true);
	}
	public CDBAR(long data) {
		super(data, false);
	}
	public CDBAR(long data, boolean owned) {
		super(data, owned);
	}
	
	/* (non-Javadoc)
	 * @see com.darwino.domino.napi.struct.ODSStruct#getODSType()
	 */
	@Override
	public ODSType getODSType() {
		return ODSType.CDBAR;
	}
	
	/* (non-Javadoc)
	 * @see com.darwino.domino.napi.struct.cd.CDStruct#loadVariableData(com.darwino.domino.napi.DominoAPI, long)
	 */
	@Override
	protected void _readODSVariableData(DominoAPI api, long dataPtr) {
		long ptr = dataPtr;
		
		Set<BARREC> flags = getFlags();
		if(flags.contains(BARREC.HAS_COLOR)) {
			this.color = C.getWORD(ptr, 0);
			ptr = C.ptrAdd(ptr, C.sizeOfWORD);
		} else {
			this.color = -1;
		}
		
		if(flags.contains(BARREC.ISFORMULA)) {
			// TODO handle this specially
		}
		
		int textLength = getHeader().getLength() - api.ODSLength(getODSType().getValue()) - (this.color > -1 ? C.sizeOfWORD : 0);
		if(textLength == 0) {
			this.text = ""; //$NON-NLS-1$
		} else {
			String text = C.getLMBCSString(ptr, 0, textLength);
			this.text = text;
		}
	}
	
	/* ******************************************************************************
	 * Struct field getters/setters
	 ********************************************************************************/

	@Override
	public WSIG getHeader() {
		_checkRefValidity();
		return new WSIG(C.ptrAdd(data, _Header));
	}
	@Override
	public void setHeader(WSIG signature) {
		_checkRefValidity();
		C.memcpy(data, _Header, signature.getDataPtr(), 0, WSIG.sizeOf);
	}
	
	public int getFlagsRaw() {
		return _getDWORD(_Flags);
	}
	
	public Set<BARREC> getFlags() {
		return DominoEnumUtil.valuesOf(BARREC.class, getFlagsRaw());
	}
	
	public void setFlagsRaw(int flags) {
		_setDWORD(_Flags, flags);
	}
	
	public void setFlags(Collection<BARREC> flags) {
		setFlagsRaw(DominoEnumUtil.toBitField(BARREC.class, flags));
	}
	
	/* ******************************************************************************
	 * Variable data
	 ********************************************************************************/
	
	public String getText() {
		return text;
	}
	
	/* ******************************************************************************
	 * Misc.
	 ********************************************************************************/
	
	/* (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString() {
    	if(isRefValid()) {
	    	return StringUtil.format("[{0}: Header={1}, Text={2}]", //$NON-NLS-1$
	    			getClass().getSimpleName(),
	    			getHeader(),
	    			getText()
	    	);
    	} else {
    		return super.toString();
    	}
    }

}
